"""
.. include:: ../README.md
"""

from .chromium import Chrome, chrome  # noqa: F401
from .session import CachedSession
from .utils import soups

__all__ = [
    "CachedSession",
    "chrome",
    "soups",
]



import os
import platform
import shutil
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Union

from selenium.webdriver import Chrome
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.chrome.service import Service

try:
    from chromedriver_py import binary_path as __chromedriver
except ModuleNotFoundError:
    __chromedriver = None


def find_chromedriver() -> Union[str, None]:
    """Find the ChromeDriver executable in the system PATH.

    This function determines the appropriate executable name based on the current
    operating system and searches for it in the system PATH.

    Returns:
        str: The path to the ChromeDriver executable if found, otherwise None.
    """

    if platform.system() == "Windows":
        driver = "chromedriver.exe"
    else:
        driver = "chromedriver"

    return shutil.which(driver)


def chromedriver() -> Path:
    """Get the path to the ChromeDriver executable.

    Either from package `chromedriver-py` or check System PATH.

    Returns:
        Path: The path to the ChromeDriver executable.
    """

    path = __chromedriver or find_chromedriver()

    if path is None:
        raise RuntimeError(
            "\n".join(
                (
                    "'chromedriver' could not be found in PATH.",
                    f"\tpip install {__package__}[driver]",
                )
            )
        )
    return Path(path).resolve(strict=True)


@contextmanager
def chrome(
    *,
    mobile: bool = False,
    verbose: bool = True,
    driver: Optional[os.PathLike] = None,
) -> Generator[Chrome, None, None]:
    """Create a Selenium Chrome webdriver context.

    This context manager sets up the Chrome webdriver with specified configuration
    options. It allows running in headless mode and emulating mobile devices, and
    ensures that the webdriver is properly terminated after use.

    Args:
        mobile (bool): Enable mobile emulation if True. Defaults to False.
        verbose (bool): If False, runs the browser in headless mode with a fixed window
            size. Defaults to True.
        driver (str): The path to the ChromeDriver executable.

    Yields:
        webdriver.Chrome: A configured Chrome webdriver instance.
    """

    driver = driver or chromedriver()

    browser_args = [
        "--disable-search-engine-choice-screen",
        "--no-default-browser-check",
        "--no-first-run",
        "--disable-default-apps",
        "--disable-popup-blocking",
        "--disable-extensions",
        "--disable-notifications",
    ]

    if verbose is False:
        browser_args += [
            "--headless",
            "--window-size=1920,1080",
        ]

    chrome_options = Options()
    for arg in browser_args:
        chrome_options.add_argument(arg)

    if mobile is True:
        # Set up device emulation for iPhone SE
        mobile_emulation = {
            "deviceMetrics": {"width": 375, "height": 667, "pixelRatio": 2.0},
            "userAgent": (
                "Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) "
                "AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0 Mobile/14E304 "
                "Safari/602.1"
            ),
        }
        chrome_options.add_experimental_option("mobileEmulation", mobile_emulation)

    service = Service(
        executable_path=driver,
    )
    drv = Chrome(
        service=service,
        options=chrome_options,
    )

    try:
        yield drv
    finally:
        drv.quit()



import functools
from typing import Optional, Union

import requests
import requests_cache
import responses
from url_normalize import url_normalize  # type: ignore[import]

from .chromium import Chrome


class CachedSession(requests_cache.CachedSession):

    def __contains__(self, url: str) -> bool:
        """Determine if a URL is present in the cache.

        Args:
            url (str): The URL to check.

        Returns:
            bool: True if a cached response exists for the URL, False otherwise.
        """
        return bool(self.get_response(url))

    @staticmethod
    def normalize(url: str) -> str:
        """Normalize a URL.

        Args:
            url (str): The URL to normalize.

        Returns:
            str: The normalized URL.
        """
        return url_normalize(url)

    @functools.lru_cache
    def create_key(self, url: str) -> str:
        """Create a cache key for a given URL.

        Args:
            url (str): The URL for which to generate a cache key.

        Returns:
            str: The created cache key.
        """
        response = requests.Request("GET", self.normalize(url)).prepare()
        return self.cache.create_key(response)

    def urls(self, **kwargs) -> list[str]:
        """Retrieve a list of cached URLs.

        Args:
            **kwargs: Additional keyword arguments for filtering the URLs.

        Returns:
            List[str]: A list of URLs present in the cache.
        """
        return self.cache.urls(**kwargs)

    def get_response(self, url: str, **kwargs) -> Union[requests.Response, None]:
        """Get the cached response for the specified URL.

        Args:
            url (str): The URL to look up.
            **kwargs: Additional keyword arguments.

        Returns:
            requests.Response: The cached HTTP response or None if not found.
        """
        key = self.create_key(url)
        return self.cache.get_response(key, **kwargs)

    def save_response(self, response: requests.Response, **kwargs) -> None:
        """Save an HTTP response to the cache.

        Args:
            response (requests.Response): The HTTP response to cache.
            **kwargs: Additional keyword arguments.
        """
        self.cache.save_response(response, **kwargs)

    def save_driver(self, driver: Chrome, **kwargs) -> requests.Response:
        """Save a response generated from a WebDriver's page source to the cache.

        Args:
            driver (WebDriver): The WebDriver instance.
            **kwargs: Additional keyword arguments.
        """
        response = self.response(driver)
        self.save_response(response, **kwargs)

        return response

    def save(
        self, response: Union[requests.Response, Chrome], **kwargs
    ) -> requests.Response:
        """Save an HTTP or WebDriver response to the cache.

        Args:
            response (Union[requests.Response, WebDriver]): The HTTP response or
                WebDriver instance to cache.
            **kwargs: Additional keyword arguments.

        Returns:
            requests.Response: The cached HTTP response.
        """

        if isinstance(response, Chrome):
            return self.save_driver(response, **kwargs)

        self.save_response(response, **kwargs)
        return response

    def get(self, url: str, **kwargs) -> requests_cache.AnyResponse:  # type: ignore[override] # noqa: E501
        """Perform a GET request with a normalized URL.

        Args:
            url (str): The URL to request.
            **kwargs: Additional keyword arguments passed to the GET request.

        Returns:
            requests.Response: The HTTP response.
        """
        return super().get(self.normalize(url), **kwargs)

    @classmethod
    def response(
        cls,
        driver: Chrome,
        *,
        encoding: Optional[str] = None,
    ) -> requests.Response:
        """Generate an HTTP response from a WebDriver's current page.

        Args:
            driver (WebDriver): The WebDriver instance.
            encoding (str, optional): The encoding for the page source. Defaults to
                "utf-8" if not specified.

        Returns:
            requests.Response: The generated HTTP response.
        """
        encoding = encoding or "utf-8"
        url = cls.normalize(driver.current_url)
        body = driver.page_source.encode(encoding)

        # Step 1: Mock the request using the responses library
        with responses.RequestsMock() as r:
            r.add(
                responses.GET,
                url=url,
                body=body,
                status=200,
                content_type=f"text/html; charset={encoding}",
            )

            # Step 3: Make the mocked request
            response = requests.get(url)

        return response



from collections.abc import Generator, Iterable
from typing import Callable, Optional

from .chromium import chrome
from .session import CachedSession, Chrome

try:
    from bs4 import BeautifulSoup
except ModuleNotFoundError as e:

    class BeautifulSoup:  # type: ignore[no-redef]
        __error = e

        def __new__(cls, *args, **kwargs):
            raise RuntimeError(
                "\n".join(
                    (
                        f"{cls.__error}",
                        f"\tpip install {__package__}[bs4]",
                    )
                )
            ) from cls.__error


def soups(
    urls: Iterable[str],
    *,
    scraper: Optional[Callable[[Chrome], None]] = None,
    **kwargs,
) -> Generator[BeautifulSoup, None, None]:
    """
    Generate BeautifulSoup objects for the given URLs using a cached session with
    Selenium.

    This function fetches the content of each URL using a cached session to avoid
    redundant network calls. If a cached response is not found for a URL, it will load
    the URL with Selenium, optionally process it with a scraper callback, and then cache
    the response. Finally, the function yields a BeautifulSoup object parsed from the
    response content.

    Args:
        urls (Iterable[str]): An iterable of URL strings to be fetched.
        scraper (Optional[Callable[[Chrome], None]]): An optional callback that
            accepts a Chorme instance for additional page processing before caching.
        **kwargs: Additional keyword arguments that are passed to the CachedSession.

    Returns:
        Generator[BeautifulSoup, None, None]: A generator yielding BeautifulSoup objects
            created from the HTML content of each fetched URL.
    """

    # extract kwargs for chrome()
    options = {
        k: kwargs.pop(k)
        for k in (
            "mobile",
            "verbose",
            "driver",
        )
        if k in kwargs.keys()
    }

    with CachedSession(**kwargs) as session:
        with chrome(**options) as driver:

            for url in urls:
                response = session.get_response(url)

                if response is None:
                    driver.get(url)

                    if scraper is not None:
                        scraper(driver)

                    response = session.save_driver(driver)

                yield BeautifulSoup(response.content, "html.parser")
